import csv

from typing import Generator, Optional, List, Tuple

from sklearn.preprocessing import LabelBinarizer

from dostoevsky.tokenization import BaseTokenizer


class BaseCorpusContainer:

    def get_prepared_data(self) -> Generator[
        Tuple[List[List[float]], List[int]], None, None
    ]:
        raise NotImplementedError


class RusentimentCorpus(BaseCorpusContainer):

    CSV_DELIMITER: str = ','
    CSV_QUOTECHAR: str = '"'

    UNKNOWN_LABEL: str = 'unknown'

    LABELS: List[str] = [
        'positive',
        'negative',
        'neutral',
        'skip',
        'speech',
        UNKNOWN_LABEL,
    ]

    def __init__(
        self,
        data_path: Optional[str],
        tokenizer: BaseTokenizer,
        lemmatize: bool = True,
    ):
        self.data_path = data_path
        self.tokenizer = tokenizer
        self.lemmatize = lemmatize
        self.label_encoder = self.get_label_encoder()

    def get_label_encoder(self) -> LabelBinarizer:
        label_encoder = LabelBinarizer()
        return label_encoder.fit(self.LABELS)

    def get_prepared_data(self) -> Generator[Tuple[List[List[float]], List[int]], None, None]:
        if not self.data_path:
            raise ValueError('data_path is None')
        with open(self.data_path) as source:
            reader = csv.reader(
                source,
                delimiter=self.CSV_DELIMITER,
                quotechar=self.CSV_QUOTECHAR,
            )
            for i, (label, text) in enumerate(reader):
                if i == 0:  # skip headers
                    continue
                encoded_label, *_ = self.label_encoder.transform([label])
                tokens = self.tokenizer.split(text, lemmatize=self.lemmatize)
                yield tokens, encoded_label



import os

from typing import List, Dict, Optional

from fasttext import load_model as load_fasttext_model

from dostoevsky.tokenization import BaseTokenizer
from dostoevsky.corpora import BaseCorpusContainer, RusentimentCorpus
from dostoevsky.data import DATA_BASE_PATH


class BaseModel:

    def __init__(
        self,
        tokenizer: BaseTokenizer,
        lemmatize: bool = True,
        model_path: Optional[str] = None,
        corpus: Optional[BaseCorpusContainer] = None,
    ):
        self.model_path = model_path
        self.tokenizer = tokenizer
        self.lemmatize = lemmatize
        self.corpus = corpus
        self.model = (
            self.get_compiled_model()
            if self.model_path
            else self.get_raw_model()
        )

    def get_compiled_model(self):
        raise NotImplementedError

    def preprocess_input(self, sentences: List[str]):
        raise NotImplementedError

    def predict(self, sentences: List[str]):
        raise NotImplementedError

    def get_raw_model(self):
        raise NotImplementedError

class FastTextSocialNetworkModel(BaseModel):
    '''
    FastText model trained on RuSentiment dataset.
    '''

    SENTENCE_LENGTH: Optional[int] = None

    MODEL_PATH: str = os.path.join(
        DATA_BASE_PATH,
        'models/fasttext-social-network-model.bin'
    )

    def __init__(
        self,
        tokenizer: BaseTokenizer,
        lemmatize: bool = False,
    ):
        super(FastTextSocialNetworkModel, self).__init__(
            tokenizer=tokenizer,
            lemmatize=lemmatize,
            model_path=self.MODEL_PATH,
            corpus=None
        )

    def get_compiled_model(self):
        return load_fasttext_model(self.MODEL_PATH)

    def preprocess_input(self, sentences: List[str]) -> List[str]:
        return [
            ' '.join(
                token for token, _ in self.tokenizer.split(
                    sentence, lemmatize=self.lemmatize
                )
            )
            for sentence in sentences
        ]

    def predict(self, sentences: List[str], k: int = -1) -> List[
        Dict[str, float]
    ]:
        X = self.preprocess_input(sentences)
        Y = (
            self.model.predict(sentence, k=k) for sentence in X
        )
        return [
            dict(zip(
                (label.replace('__label__', '') for label in labels),
                scores
            )) for labels, scores in Y
        ]



def test_rusentiment_corpus_get_prepared_data(
    rusentiment_corpus,
):
    X_train, y_train = [], []
    for i, (vectors, label) in enumerate(
        rusentiment_corpus.get_prepared_data()
    ):
        X_train.append(vectors)
        y_train.append(label)
        if i >= 9:
            break
    assert len(X_train) == len(y_train)
    assert X_train[0]



import os
import lzma
import typing
import tarfile
import urllib.request


DATA_BASE_PATH: str = os.path.dirname(os.path.abspath(__file__))
STORAGE_BASE_URL: str = 'https://storage.b-labs.pro/'


AVAILABLE_FILES: typing.Dict[str, typing.Tuple[str, str]] = {
    'cnn-social-network-model': ('models/cnn-social-network-model.tar.xz', 'models/cnn-social-network-model.tar.xz'),
    'fasttext-social-network-model': ('models/fasttext-social-network-model.tar.xz', 'models/fasttext-social-network-model.tar.xz')
}


class DataDownloader:

    USERAGENT: str = 'Dostoevsky / 1.0'
    CHUNK_SIZE: int = 1024 * 32

    def download(self, source: str, destination: str) -> int:
        destination_path: str = os.path.join(DATA_BASE_PATH, destination)
        url: str = os.path.join(STORAGE_BASE_URL, source)
        request = urllib.request.Request(url)
        request.add_header('User-Agent', self.USERAGENT)
        response = urllib.request.urlopen(request)
        with open(destination_path, 'wb') as output:
            filesize: int = 0
            while source:
                chunk = response.read(self.CHUNK_SIZE)
                if not chunk:
                    break
                filesize += len(chunk)
                output.write(chunk)
        # assume that we always distribute data as .tar.xz archives
        with lzma.open(destination_path) as f:
            with tarfile.open(fileobj=f) as tar:
                tar.extractall(os.path.dirname(destination_path))
        return filesize



from typing import List, Tuple, Set, Optional

from pymorphy2 import MorphAnalyzer
from razdel import tokenize as regex_tokenize
from b_labs_models import Tokenizer, POSTagger
from russian_tagsets import converters


opencorpora_to_ud_convert = converters.converter('opencorpora-int', 'ud14')


class BaseTokenizer:
    '''
    Base tokenizer interface
    '''

    def split(self, text: str, lemmatize: bool = True) -> List[
        Tuple[str, Optional[str]]
    ]:
        '''
        Each subclass of BaseTokenizer must inplement this method
        Returns list with tuples like that one:
        [('приехать', 'VERB'), ('дом', 'NOUN')]
        '''
        raise NotImplementedError


class RegexTokenizer(BaseTokenizer):
    '''
    Tokenizer based on one of most accurate and fast tokenizers for russian
    language text - razdel. This tokenizer doesn't performs POS tagging
    and lemmatization.
    '''

    def split(self, text: str, lemmatize: bool = False) -> List[
        Tuple[str, Optional[str]]
    ]:
        return [
            (token.text.lower(), None) for token in regex_tokenize(text)
        ]


class BaselineTokenizer(BaseTokenizer):
    '''
    Simple tokenizer that uses pre-trained CRF model from B-Labs and
    takes first form of each word from pymorphy2 output
    '''

    POS_WITHOUT_NORMAL_FORM: Set[str] = {
        'SYMB',
        'NUMB',
        'PNCT',
        'LATN'
    }

    def __init__(self):
        self.tokenizer = Tokenizer()
        self.tagger = POSTagger()
        self.morph = MorphAnalyzer()

    def get_token_text(self, token: str, pos: str, lemmatize: bool = False):

        if pos in self.POS_WITHOUT_NORMAL_FORM or not lemmatize:
            return token.lower()

        text: str = token
        forms: List[Tuple[str, str, str]] = [
            (form.word, form.normal_form, form.tag)
            for form in self.morph.parse(token)
        ]
        for (_, normal_form, tag) in forms:
            if pos in tag:
                text = normal_form
                break
        return text.lower()

    def split(self, text: str, lemmatize: bool = False):
        tokens: List[str] = list(self.tokenizer.split(text))
        parts_of_speech: List[str] = list(self.tagger.tag(tokens))
        return [
            (self.get_token_text(token, pos, lemmatize), pos)
            for token, pos in zip(tokens, parts_of_speech)
        ]


class UDBaselineTokenizer(BaselineTokenizer):
    '''
    Tokenizer that works similar to baseline tokenizer, but also
    converts OpenCorpora part of speech tags to Universal Dependencies
    format, because some word vectors (notably, RusVectores) was trained on UD tagset
    '''

    UNKNOWN_POS_TAG: str = 'X'

    def split(self, text: str, lemmatize: bool = False):
        tokens: List[Tuple[str, str]] = (
            super(UDBaselineTokenizer, self)
            .split(text=text, lemmatize=lemmatize)
        )
        results: List[Tuple[str, str]] = [
            (token, opencorpora_to_ud_convert(pos).split(' ')[0])
            if pos else (token, self.UNKNOWN_POS_TAG) for token, pos in tokens
        ]
        return results



from setuptools import (
    setup,
    find_packages,
)


def get_long_description() -> str:
    with open('README.md') as source:
        return source.read()


setup(
    name='dostoevsky',
    version='0.4.0',
    description='Sentiment analysis library for russian language',
    url='https://github.com/bureaucratic-labs/dostoevsky',
    author='Bureaucratic Labs',
    author_email='hello@b-labs.pro',
    license='MIT',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing :: Linguistic',
    ],
    keywords='natural language processing, sentiment analysis',
    packages=find_packages(),
    package_data={
        'dostoevsky': [
            'data/corpora/*',
            'data/embeddings/*',
            'data/models/*',
        ]
    },
    scripts=['bin/dostoevsky'],
    install_requires=[
        'b-labs-models == 2017.8.22',
        'razdel==0.4.0',
        'fasttext==0.9.1',
        'pymorphy2==0.8',
        'pytest==5.2.4',
        'russian-tagsets==0.6',
        'scikit-learn==0.21.3',
    ],
)



from typing import List, Tuple, Optional


def test_baseline_tokenizer_base_case(baseline_tokenizer):
    tokens: List[
        Tuple[str, str]
    ] = baseline_tokenizer.split('он приехал домой')
    assert tokens == [
        ('он', 'NPRO'),
        ('приехал', 'VERB'),
        ('домой', 'ADVB'),
    ]


def test_baseline_tokenizer_with_lemmatization(baseline_tokenizer):
    tokens: List[
        Tuple[str, str]
    ] = baseline_tokenizer.split('он приехал домой', lemmatize=True)
    assert tokens == [
        ('он', 'NPRO'),
        ('приехать', 'VERB'),
        ('домой', 'ADVB'),
    ]


def test_baseline_tokenizer_with_lemmatization_and_mixed_case_text(baseline_tokenizer):
    tokens: List[
        Tuple[str, str]
    ] = baseline_tokenizer.split('оН пРиехал доМой', lemmatize=True)
    assert tokens == [
        ('он', 'NPRO'),
        ('приехать', 'VERB'),
        ('домой', 'ADVB'),
    ]


def test_ud_baseline_tokenizer_simple_case(ud_baseline_tokenizer):
    tokens: List[
        Tuple[str, str]
    ] = ud_baseline_tokenizer.split('она была очень красивой')
    assert tokens == [
        ('она', 'PRON'),
        ('была', 'VERB'),
        ('очень', 'ADV'),
        ('красивой', 'ADJ'),
    ]


def test_ud_baseline_tokenizer_with_special_words(ud_baseline_tokenizer):
    tokens: List[
        Tuple[str, str]
    ] = ud_baseline_tokenizer.split('123, test')
    assert tokens == [
        ('123', 'NUM'),
        (',', 'PUNCT'),
        ('test', 'X'),
    ]


def test_ud_baseline_tokenizer_with_unknown_word(ud_baseline_tokenizer):
    tokens: List[
        Tuple[str, str]
    ] = ud_baseline_tokenizer.split('asdasdasde21e')
    assert tokens == [
        ('asdasdasde21e', 'X'),
    ]


def test_regex_tokenizer_base_case(regex_tokenizer):
    tokens: List[
        Tuple[str, Optional[str]]
    ] = regex_tokenizer.split('он приехал домой')
    assert tokens == [
        ('он', None),
        ('приехал', None),
        ('домой', None),
    ]


def test_regex_tokenizer_lower_case(regex_tokenizer):
    tokens: List[
        Tuple[str, Optional[str]]
    ] = regex_tokenizer.split('оН пРиехал доМой')
    assert tokens == [
        ('он', None),
        ('приехал', None),
        ('домой', None),
    ]



import os
import pytest


from dostoevsky.tokenization import (
    BaselineTokenizer,
    UDBaselineTokenizer,
    RegexTokenizer,
)
from dostoevsky.corpora import RusentimentCorpus
from dostoevsky.data import DataDownloader, DATA_BASE_PATH


@pytest.fixture
def baseline_tokenizer():
    return BaselineTokenizer()


@pytest.fixture
def ud_baseline_tokenizer():
    return UDBaselineTokenizer()


@pytest.fixture
def regex_tokenizer():
    return RegexTokenizer()


@pytest.fixture(scope='session')
def data_downloader():
    return DataDownloader()


@pytest.fixture(scope='session')
def embeddings_path(data_downloader) -> str:
    filesize: int = data_downloader.download(
        'embeddings/vk-min-100-300d-none-limited.tar.xz',
        'embeddings/vk-min-100-300d-none-limited.tar.xz',
    )
    assert filesize > 0
    return os.path.join(
        DATA_BASE_PATH,
        'embeddings/vk-min-100-300d-none-limited.vec',
    )


@pytest.fixture(scope='session')
def rusentiment_corpus_data(data_downloader):
    filesize: int = data_downloader.download(
        'corpora/rusentiment.tar.xz',
        'corpora/rusentiment',
    )
    assert filesize > 0
    return os.path.join(
        DATA_BASE_PATH,
        'corpora/',
    )


@pytest.fixture(scope='session')
def rusentiment_corpus_path(rusentiment_corpus_data):
    return os.path.join(rusentiment_corpus_data, 'rusentiment_random_posts.csv')


@pytest.fixture(scope='session')
def rusentiment_test_corpus_path(rusentiment_corpus_data):
    return os.path.join(rusentiment_corpus_data, 'rusentiment_test.csv')


@pytest.fixture(scope='session')
def rusentiment_baseline_tokenizer():
    return UDBaselineTokenizer()


@pytest.fixture(scope='session')
def rusentiment_corpus(
    rusentiment_corpus_path,
    rusentiment_baseline_tokenizer,
):
    return RusentimentCorpus(
        data_path=rusentiment_corpus_path,
        tokenizer=rusentiment_baseline_tokenizer,
    )


@pytest.fixture(scope='session')
def rusentiment_test_corpus(
    rusentiment_test_corpus_path,
    rusentiment_baseline_tokenizer,
):
    return RusentimentCorpus(
        data_path=rusentiment_test_corpus_path,
        tokenizer=rusentiment_baseline_tokenizer,
    )
